# -*- coding: utf-8 -*-
"""
In a nutshell, we have receiving module which is responsible for converting
incoming message to a Callback object; then we have handling module to handle
internal bookkeeping for Messaging object (which is inside Callback); lastly
we use sender module to send something back.

But we need a mechanism to glue them together so that the system knows which
sender(s) should be invoked when we received some kinds of Messaging
"""
import importlib
import logging

from djmessenger.utils.serializable import Serializable
from djmessenger.utils.utils import load_class
from djmessenger.settings import DJM_ROUTING_POLICY
from djmessenger.handling import UserProfileHandler


logger = logging.getLogger(__name__)


class TargetClass(Serializable):
    def __init__(self, name, args=dict()):
        self.name = name
        self.args = args

    def get_class(self):
        return load_class(self.name)

    def get_args(self):
        # args could be none if the policy file does not specify it, which is
        # valid
        return getattr(self, 'args', {})

    def __repr__(self):
        return 'class %s with args %s' % (self.name, getattr(self, 'args', {}))

    def __str__(self):
        return repr(self)


class TargetSenderClass(TargetClass):
    pass


class TargetHandlerClass(TargetClass):
    custom_obj_map = {
        'senders': [TargetSenderClass, list]
    }

    def __init__(self, name, args=list(), senders=list()):
        super().__init__(name, args)
        self.senders = senders


class Router(Serializable):
    custom_obj_map = {
        'handlers': [TargetHandlerClass, list]
    }

    def __init__(self, ttype, handlers=list()):
        from djmessenger.receiving import ReceivingType

        self.type = ReceivingType.value_of(ttype)
        self.handlers = handlers

    def get_receiving_type(self):
        from djmessenger.receiving import ReceivingType

        try:
            ret = ReceivingType.value_of(self.type)
            return ret
        except KeyError as e:
            logger.error('Given type %s is not a valid receiving type [%s]' %
                         (self.type, ReceivingType.members().keys()))
            raise e


class Policy(Serializable):
    custom_obj_map = {
        'routers': [Router, list]
    }

    def __init__(self, routers=list()):
        self.routers = routers

    def get_router(self, rtype):
        for router in self.routers:
            if router.get_receiving_type() == rtype:
                return router
        return None

    def get_handlers(self, rtype):
        """
        Given a ReceivingType, get all applicable TargetHandlerClass classes

        @param rtype:
        @return:
        @rtype: list of TargetHandlerClass
        """
        from djmessenger.receiving import ReceivingType

        ret = []
        if self.get_router(ReceivingType.DEFAULT):
            for handler in self.get_router(ReceivingType.DEFAULT).handlers:
                if issubclass(handler.get_class(), UserProfileHandler):
                    ret.insert(0, handler)
                else:
                    ret.append(handler)
        if self.get_router(rtype) is not None:
            for handler in self.get_router(rtype).handlers:
                if issubclass(handler.get_class(), UserProfileHandler):
                    ret.insert(handler, 0)
                else:
                    ret.append(handler)
        return ret

    def get_senders(self, rtype, handler_class):
        for handler in self.get_handlers(rtype):
            if handler_class == handler:
                return handler.senders


try:
    policy = Policy.deserialize(DJM_ROUTING_POLICY)
except Exception as e:
    logger.error('Failed to load policy file from setting, please double check'
                 'on your settings for DJM_ROUTING_POLICY')


if __name__ == '__main__':
    print(policy)



# -*- coding: utf-8 -*-
from django.test import TestCase
from djmessenger.routing import *
import os
from djmessenger.settings import *
import json
from pprint import pprint
from djmessenger.handling import BaseHandler
from djmessenger.receiving import ReceivingType


class TestRouter(TestCase):
    def test_deserialization(self):
        file = os.path.join(DJM_BASE_DIR, 'ref', 'default_routing.json')
        self.assertTrue(os.path.exists(file))
        policy = Policy.read_json(file)
        # pprint(router)
        # print(ReceivingType.members())
        for router in policy.routers:
            self.assertTrue(isinstance(router, Router))
            try:
                # print(type(router.get_receiving_type()))
                self.assertTrue(isinstance(router.get_receiving_type(), ReceivingType))
            except Exception as e:
                self.fail('should not fail but %s' % e)
            self.assertTrue(isinstance(router.handlers, list))
            for handler in router.handlers:
                self.assertTrue(isinstance(handler, TargetHandlerClass))
                try:
                    self.assertTrue(issubclass(handler.get_class(), BaseHandler))
                except Exception as e:
                    self.fail('should not fail but %s' % e)

    def test_basic(self):
        # policy has been loaded
        self.assertEqual(4, len(policy.routers))
        self.assertEqual(ReceivingType.DEFAULT, policy.get_router(ReceivingType.DEFAULT).get_receiving_type())
        handlers = policy.get_handlers(ReceivingType.DEFAULT)
        self.assertEqual(2, len(handlers))
        for handler in handlers:
            self.assertTrue(issubclass(handler, BaseHandler), 'handler is type %s' % type(handler))



# -*- coding: utf-8 -*-
"""
# Handling module

Handling module provides abstract class and some implementation for handling a
`receiving.Messaging` object

## BaseHandler

Defines `should_handle()` and `handle()` which need to be overridden by
subclasses

In `should_handle()`, you look at the `Messaging` object and determine whether
this handler should handle this `Messaging`

In `handle()`, you just perform operations on the data provided by `Messaging`

## Provided Handlers

### DummyHandler

`should_handle()` just passes without checking anything

`handle()` does nothing

This handler is for routing policy to route a `Messaging` directly to senders

### LocationHandler

`should_handle()` if the ReceivingType is LOCATION and we have location data

`handle()` save the GPS coordinates into database

### BasePayloadHandler (TBD)

This is an ABC that defines `get_payload()`, only `ReceivingType.POSTBACK` and
 `ReceivingType.QUICK_REPLY` would have `payload`

### BasePostbackPayloadHandler (TBD)

This is an ABC that you should subclass to handle payload from
`ReceivingType.POSTBACK`

### BaseQuickReplyPayloadHandler (TBD)

This is an ABC that you should subclass to handle payload from
`ReceivingType.QUICK_REPLY`

### SaveMessagingHandler

`should_handle()`: if DJM_SAVE_MESSAGING settings is True

`handle()`: save the `Messaging` object into database

### SimpleTextBaseHandler

`should_handle()`: if the `Messaging` contains a text; if regex was provided,
text must match the regex

`handle()`: do nothing. You should subclass this class if you want to do
anything to the received SIMPLE_TEXT

### ThumbUpHandler

`should_handle()`: if we think the `Messaging` is a sticker and it is a thumb up

`handle()`: increment the thumbup count for the psid in the database

### UserProfileHandler

`should_handle()`: pass

`handle()`: if DJM_SAVE_USER_PROFILE was True, fetch user details using graph
api and save it; otherwise only save the psid

> This handler should be always in DEFAULT section of the policy, otherwise
  you won't be able to send anything back since you won't have user's psid
"""
import json
import logging
from abc import ABC, abstractmethod

import requests

from djmessenger.models import UserLocation, FBUserProfile
from djmessenger.payload import Payload
from djmessenger.receiving import Messaging, ReceivingType
from djmessenger.utils.utils import get_class_name, load_class

logger = logging.getLogger(__name__)


class BaseHandler(ABC):
    """
    BaseHandler is an abstract base class to

    1. Determine whether the given messaging object should be handled by this
       handler
    2. Actually handle it by adding something to database, or other internal
       work

    Each received messaging object will be applied to all defined handlers,
    which means you can actually do multiple stuff for a single messaging as
    long as your should_handle returns True
    """
    def __init__(self, messaging):
        """

        @param messaging:
        @type messaging: Messaging
        """
        self.messaging = messaging

    def get_psid(self):
        return self.messaging.get_psid()

    @abstractmethod
    def should_handle(self):
        """
        Whether the given messaging is applicable for this handler

        @return: True if the messaging should be handled; False otherwise
        @rtype: bool
        """
        pass

    @abstractmethod
    def handle(self):
        """
        Actually handles the messaging

        @return:
        """
        pass

    def __repr__(self):
        return self.__class__.__name__

    def __str__(self):
        return repr(self)


class SaveMessagingHandler(BaseHandler):
    """
    This handler saves the messaging into the database for further reference
    """
    def should_handle(self):
        return True

    def handle(self):
        logger.debug('SaveMessagingHandler is saving messaging...')
        from djmessenger.models import Messaging as ModelMessaging
        ModelMessaging.objects.create(body=self.messaging.serialize(),
                                      type=self.messaging.get_receiving_type())
        logger.debug('SaveMessagingHandler successfully saved messaging')


class UserProfileHandler(BaseHandler):
    """
    Every messaging has sender.id, based on settings.DJM_SAVE_USER_PROFILE, this
    handler save user profile into database using models.FBUserProfile.

    If DJM_SAVE_USER_PROFILE was True, we fetch user profile using graph API;
    otherwise we only save user psid to the database
    """
    def should_handle(self):
        return True

    def handle(self):
        from djmessenger.models import FBUserProfile
        from djmessenger.settings import DJM_USER_DETAILS_URL, DJM_PAGE_ACCESS_TOKEN, DJM_SAVE_USER_PROFILE

        # we already checked this id exists in should_handle
        psid = self.get_psid()
        logger.debug('UserProfileHandler is handling messaging from user %s' % psid)
        try:
            FBUserProfile.objects.get(pk=psid)
            # user already exists
            logger.debug('PSID %s already exists, no need to fetch' % psid)
        except FBUserProfile.DoesNotExist:
            logger.debug('User %s does not exist, trying to create' % psid)
            if DJM_SAVE_USER_PROFILE:
                # user does not exist
                logger.debug('Because DJM_SAVE_USER_PROFILE was True, Ready to '
                             'fetch and save user details for PSID %s' % psid)
                status = requests.get(
                    DJM_USER_DETAILS_URL % psid,
                    {'access_token': DJM_PAGE_ACCESS_TOKEN}
                )
                if status.status_code != 200:
                    logger.error('Failed to fetch user details using Facebook Graph'
                                 'API for PSID %s' % psid)
                else:
                    user_detail = status.json()
                    logger.debug('Successfully fetched user profile for psid %s'
                                 ': %s'
                                 % (psid, user_detail))
                    try:
                        fp = FBUserProfile(**user_detail)
                        fp.psid = psid
                        fp.save()
                        logger.debug('Successfully handled creating user '
                                     'profile for %s' % psid)
                    except:
                        logger.debug('Failed to create FBUserProfile from user'
                                     'details: %s' % user_detail)
            else:
                # do not fetch user profile, only save psid
                logger.debug('Because DJM_SAVE_USER_PROFILE was False, only '
                             'save user PSID')
                FBUserProfile.objects.create(psid=psid)


class LocationHandler(BaseHandler):
    """
    If the user sends a location to the BOT (by click on the map pin icon next
    to thumb up), this handler saves this coordinates to the database
    """
    def should_handle(self):
        logger.debug('Ready to determine whether %s should handle this '
                     'messaging' % self.__class__.__name__)
        ret = self.messaging.get_receiving_type() == ReceivingType.LOCATION
        logger.debug('We check whether the messaging callback type is '
                     'MESSAGE_RECEIVED_LOCATION. The result is %s' % ret)
        return ret

    def handle(self):
        message = self.messaging.message
        psid = self.get_psid()
        logger.debug(
            'LocationHandler is handling messaging from user %s' % psid)
        try:
            user = FBUserProfile.objects.get(pk=psid)
            timestamp = self.messaging.timestamp if hasattr(self.messaging,
                                                            'timestamp') else None
            for atta in self.messaging.message.attachments:
                if not atta.type == 'location':
                    continue
                location = UserLocation(user=user,
                                        latitude=atta.payload.coordinates.lat,
                                        longitude=atta.payload.coordinates.long,
                                        timestamp=timestamp,
                                        mid=self.messaging.message.mid,
                                        seq=self.messaging.message.seq,
                                        url=getattr(atta, 'url', None)
                                        )
                location.save()
            logger.debug(
                'Successfully handled message containing location sent '
                'from %s' % psid)
        except FBUserProfile.DoesNotExist:
            logger.debug('No profile for psid %s, it is probably because'
                         'UserProfileHandler was not enabled.' % psid)


class ThumbUpHandler(BaseHandler):
    """
    Handles when the user sends a thumb up
    """
    def should_handle(self):
        logger.debug('Ready to determine whether %s should handle this '
                     'messaging' % self.__class__.__name__)
        ret = self.messaging.get_receiving_type() == ReceivingType.STICKER \
            and str(self.messaging.get_sticker_id())[-3:] in ('810', '814', '822')
        logger.debug('We determine whether we should handle the messaging by '
                     '1. check if the callback type is MESSAGE_RECEIVED_STICKER'
                     ', 2. check if the last 3 digit of sticker id is either '
                     '810, 814 or 822. The result is %s' % ret)
        return ret

    def handle(self):
        psid = self.get_psid()
        logger.debug('ThumbUpHandler is handling Thumbup from %s' % psid)
        try:
            user = FBUserProfile.objects.get(pk=psid)
            user.thumbups += 1
            user.save()
        except FBUserProfile.DoesNotExist:
            logger.debug('No profile for psid %s, it is probably because'
                         'UserProfileHandler was not enabled.' % psid)


class DummyHandler(BaseHandler):
    """
    This handler does nothing, just pass
    """
    def should_handle(self):
        logger.debug('DummyHandler just passes')
        return True

    def handle(self):
        pass


# TODO: how to handle postback?
class BasePayloadHandler(BaseHandler):
    """
    Postback and QuickReply contains payload, and in order to support request
    chaining, we need to provide a default base class for Postback and
    QuickReply.

    Since the payload in Postback and QuickReply is merely a simple string
    limited to 1000 chars, we can utilize this space to send some bookkeeping
    info to achieve request chaining.

    We are going to make the payload (which is a plain text) looks like a valid
    json object so that we can deserialize it back to a dict and then we can
    figure out which handler was the sender, then do corresponding actions
    """

    @abstractmethod
    def get_payload_string_from_messaging(self):
        """

        @return: the corresponding payload string from Messaging
        @rtype: str
        """
        pass

    def get_payload(self):
        """
        Payload class must be a subclass of djmessenger.utils.payload.Payload,
        and the class defines an attribute `payload_class`, we can simply load
        the class from this fully qualified class name

        @return: an instance of the subclass of Payload
        @rtype: an instance of the subclass of Payload
        """
        payload_string = self.get_payload_string_from_messaging()
        try:
            return Payload.get_instance(payload_string)
        except Exception as e:
            logger.error('payload string %s should be a serialization of a '
                         'subclass of Payload, but failed to determine the '
                         'subclass and load it because %s' % (payload_string, e))


class BasePostbackHandler(BasePayloadHandler):

    def should_handle(self):
        return self.messaging.get_receiving_type() == ReceivingType.POSTBACK \
               and self.get_payload()

    def get_payload_string_from_messaging(self):
        """

        @return: The payload portion from Messaging
        @rtype: str
        """
        return self.messaging.get_postback_payload()

    @abstractmethod
    def handle(self):
        pass


class BaseQuickReplyHandler(BasePayloadHandler):
    def should_handle(self):
        return self.messaging.get_receiving_type() == ReceivingType.QUICK_REPLY \
               and self.get_payload()

    def get_payload_string_from_messaging(self):
        return self.messaging.get_quick_reply_payload()

    def get_text(self):
        """
        quick reply will come with text which is the quick reply title

        @return:
        """
        return self.messaging.get_text()

    @abstractmethod
    def handle(self):
        pass


class SimpleTextBaseHandler(BaseHandler):
    def __init__(self, messaging, regex=None):
        super().__init__(messaging)
        self.regex = regex

    def should_handle(self):
        from re import compile

        if not self.regex:
            ret = len(self.messaging.get_text()) > 0
            logger.debug('No regex provided for SimpleTextBaseHandler, '
                         'should_handle() evaluates as %s because text [%s] was'
                         ' provided' % (ret, self.messaging.get_text()))
            return ret
        rex = compile(self.regex)
        res = rex.fullmatch(self.get_text())
        ret = res and len(self.messaging.get_text()) > 0
        logging.debug('Given text and regex (%s, %s), should_handle() is %s' %
                      (self.get_text(), self.regex, ret))
        return ret

    def get_text(self):
        return self.messaging.get_text()

    def handle(self):
        pass


class MultimediaBaseHandler(BaseHandler):
    def should_handle(self):
        receiving_type = self.messaging.get_receiving_type()
        return receiving_type in (ReceivingType.IMAGE, ReceivingType.AUDIO,
                                  ReceivingType.VIDEO, ReceivingType.FILE)

    @abstractmethod
    def handle(self):
        pass


class AuthenticationBaseHandler(BaseHandler):
    def should_handle(self):
        return self.messaging.get_receiving_type() == ReceivingType.AUTHENTICATION

    @abstractmethod
    def handle(self):
        pass


class AccountLinkingBaseHandler(BaseHandler):
    # TODO
    def should_handle(self):
        pass

    @abstractmethod
    def handle(self):
        pass


class DeliveryBaseHandler(BaseHandler):
    # TODO
    def should_handle(self):
        pass

    @abstractmethod
    def handle(self):
        pass


class ReadBaseHandler(BaseHandler):
    # TODO
    def should_handle(self):
        pass

    @abstractmethod
    def handle(self):
        pass


class EchoBaseHandler(BaseHandler):
    # TODO
    def should_handle(self):
        pass

    @abstractmethod
    def handle(self):
        pass



# -*- coding: utf-8 -*-
from django.utils.translation import ugettext as _
from re import escape

"""
# Routing Policy Explained

## Source

djmessenger/utils/default_routing_policy.py

## How routing in DJM works

1. DJM receives a payload from Facebook which contains possibly multiple
   `receiving.Messaging` object. Each `Messaging` object, we determine its
   `ReceivingType`:

   ```
   SIMPLE_TEXT
   QUICK_REPLY
   IMAGE
   AUDIO
   VIDEO
   FILE
   LOCATION
   STICKER
   POSTBACK
   AUTHENTICATION
   ACCOUNT_LINKING
   DELIVERED
   READ
   ECHO
   ```

2. In the policy, routers contains multiple router, each router defines which
   `ReceivingType` it handles.

   > Please note that 1 `ReceivingType` only has 1 router, if you defined multiple
     item for the same `ReceivingType`, only the first one will take effect

3. Each `ReceivingType` defines the Handlers (`djmessenger.handling` module) that
   should handle this `ReceivingType`.

   > You can provide args for the handler class's constructor as keyword
     arguments

4. For each Handler, you define multiple Senders (`djmessenger.sending` module)
   that will take care of sending something back to the user

   > You can provide args for the sender class's constructor as keyword
     arguments

   > The senders will be invoked iff the handler passes and was handled, if the
     `Messaging` does not meet Handler's handling prerequisite, it hence won't
     be handled and won't be sent to senders

   > You probably want to send back i18n'ed strings, so mark your sender text
     _()

       ```
       from django.utils.translation import ugettext as _
       ```

   > Be sure to use `re.escape` if your regex string contains chars that need to
     be escaped

## Example Explained

Here is a snippet in DJM's default routing policy

    ```
    {
        "type": "SIMPLE_TEXT",
        "handlers": [
            {
                "name": "djmessenger.handling.SimpleTextBaseHandler",
                "args": {
                    "regex": ""
                },
                "senders": [
                    {
                        "name": "djmessenger.sending.SimpleMessageSender",
                        "args": {
                            "text": _(
                                "Thanks for your message, we will get back to you soon")
                        }
                    }
                ]
            },
            {
                "name": "djmessenger.handling.SimpleTextBaseHandler",
                "args": {
                    "regex": "^你好$"
                },
                "senders": [
                    {
                        "name": "djmessenger.sending.SimpleMessageSender",
                        "args": {
                            "text": "您也好"
                        }
                    }
                ]
            }
        ]
    },
    ```

You can vision it very easily that:

1. A `Messaging` object that was determined as `SIMPLE_TEXT` type will be sent
   to `djmessenger.handling.SimpleTextBaseHandler` with different regex
2. The first `SimpleTextBaseHandler` has no regex arg, so it will just pass and
   the `SimpleMessageSender` simply sends back the `text`
3. The second `SimpleTextBaseHandler` has regex defined, so if the user sends
   the bot a `SIMPLE_TEXT` that matches this regex, this handler will handle it
   and then sender will be invoked

## Builtin Handlers

See Handlers for more details

"""
DJM_DEFAULT_ROUTING_POLICY = \
    {
        "routers": [
            {
                "type": "DEFAULT",
                "handlers": [
                    {
                        "name": "djmessenger.handling.SaveMessagingHandler",
                        "args": {},
                        "senders": []
                    },
                    {
                        "name": "djmessenger.handling.UserProfileHandler",
                        "args": {},
                        "senders": []
                    }
                ]
            },
            {
                "type": "STICKER",
                "handlers": [
                    {
                        "name": "djmessenger.handling.ThumbUpHandler",
                        "args": {},
                        "senders": [
                            {
                                "name": "djmessenger.sending.SimpleMessageSender",
                                "args": {
                                    "text": _("Thank you for your thumb!!!")
                                }
                            },
                            {
                                "name": "djmessenger.sending.MultimediaSender",
                                "args": {
                                    "sending_type": "image",
                                    "url": "https://dl.dropboxusercontent.com/u/717667/icons/betatesting.png"
                                }
                            }
                        ]
                    }
                ]
            },
            {
                "type": "SIMPLE_TEXT",
                "handlers": [
                    {
                        "name": "djmessenger.handling.SimpleTextBaseHandler",
                        "args": {
                            "regex": "^你好$"
                        },
                        "senders": [
                            {
                                "name": "djmessenger.sending.SimpleMessageSender",
                                "args": {
                                    "text": "您也好"
                                }
                            }
                        ]
                    },
                    {
                        "name": "djmessenger.handling.SimpleTextBaseHandler",
                        "args": {
                            "regex": "^hi$"
                        },
                        "senders": [
                            {
                                "name": "djmessenger.sending.SimpleMessageSender",
                                "args": {
                                    "text": _(
                                        "Thanks for your message, we will get back to you soon")
                                }
                            }
                        ]
                    }
                ]
            },
            {
                "type": "LOCATION",
                "handlers": [
                    {
                        "name": "djmessenger.handling.LocationHandler",
                        "args": {},
                        "senders": []
                    }
                ]
            }
        ]
    }



from __future__ import absolute_import
import unittest
import os
from djmessenger.utils.serializable import Serializable


class ChildClass1(Serializable):
    excludes = ['_hidden1']

    def __init__(self):
        self.attr1 = 1
        self.attr2 = '1'
        self.attr3 = True
        self.attr4 = (1, 11, 111)
        self.attr5 = {'key11': 1, 'key12': '12'}
        self.attr6 = 1.1
        self._hidden1 = True


class ChildClass2(ChildClass1):
    excludes = ['_hidden2', '_hidden1']

    def __init__(self):
        super(ChildClass2, self).__init__()
        self.attr4 = (2, 22, 222)
        self.attr5 = {'key21': 2, 'key22': '22'}
        self.attr6 = 2.2
        self._hidden2 = False


class ParentClass(Serializable):
    custom_obj_map = {
        'attr1': [ChildClass1, object],
        'attr2': [ChildClass2, object]
    }
    excludes = ['_ignore']

    def __init__(self):
        self.attr1 = ChildClass1()
        self.attr2 = ChildClass2()
        self.attr3 = 'parent'
        self._ignore = 'should ignore'


class ParentClass2(Serializable):
    custom_obj_map = {
        'attr1': [ParentClass, object],
        'attr2': [ChildClass1, object],
        'attr3': [ChildClass2, list]
    }

    def __init__(self):
        self.attr1 = ParentClass()
        self.attr2 = ChildClass1()
        self.attr3 = [ChildClass2(), ChildClass2()]


class TestSerializer(unittest.TestCase):
    def test_obj_map(self):
        obj = ParentClass2()
        json_string = obj.serialize()
        robj = ParentClass2.deserialize(json_string)
        self.assertTrue(isinstance(robj, ParentClass2), type(robj))
        self.assertEqual(obj.attr1.attr3, robj.attr1.attr3)
        self.assertEqual(type(robj.attr3), list)
        self.assertTrue(isinstance(robj.attr3[0], ChildClass2))
        self.assertTrue(isinstance(robj.attr2, ChildClass1))
        self.assertTrue(obj == robj)

    def test_basic(self):
        obj = ChildClass1()
        json_string = obj.serialize()
        self.assertTrue('_hidden1' not in json_string)
        robj = ChildClass1.deserialize(json_string)
        self.assertTrue('_hidden1' not in dir(robj))
        for key in vars(robj):
            self.assertTrue(key in vars(obj))
        self.assertTrue(obj.attr1, robj.attr1)

    def test_subclass(self):
        obj = ChildClass2()
        json_string = obj.serialize()
        self.assertTrue('_hidden1' not in json_string)
        self.assertTrue('_hidden2' not in json_string)
        robj = ChildClass2.deserialize(json_string)
        self.assertTrue('_hidden1' not in dir(robj))
        self.assertTrue('_hidden2' not in dir(robj))
        for key in vars(robj):
            self.assertTrue(key in vars(obj))
        self.assertTrue(obj.attr1, robj.attr1)

    def test_parent_class(self):
        obj = ParentClass()
        obj1 = obj.attr1
        obj2 = obj.attr2
        json_string_obj1 = obj1.serialize()
        json_string_obj2 = obj2.serialize()
        json_string_parent = obj.serialize()
        self.assertTrue(json_string_obj1 in json_string_parent)

        self.assertTrue(json_string_obj2 in json_string_parent)
        robj = ParentClass.deserialize(json_string_parent)
        self.assertEqual('parent', robj.attr3)

    def test_write_read(self):
        obj = ParentClass()
        file = '/tmp/test_write_read.json'
        obj.write_json(file)
        self.assertTrue(os.path.exists(file))
        robj = ParentClass.read_json(file)
        self.assertEqual('parent', robj.attr3)

    def test_json(self):
        obj = ParentClass()
        jsondict = obj.json()
        self.assertTrue(isinstance(jsondict, dict))
        robj = ParentClass.deserialize(jsondict)
        self.assertEqual(obj.attr3, robj.attr3)

if __name__ == '__main__':
    unittest.main()



import logging
import json

from django.views import generic
from django.http.response import HttpResponse
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt

from djmessenger.receiving import Callback
from djmessenger.handling import BaseHandler
from djmessenger.sending import CommonSender


logger = logging.getLogger(__name__)


class DJMBotView(generic.View):
    def get(self, request, *args, **kwargs):
        verify_token = self.request.GET.get('hub.verify_token', None)
        challenge = self.request.GET.get('hub.challenge', None)
        if verify_token and challenge:
            return HttpResponse(challenge)
        else:
            logger.error('Either verify_token [%s] or challenge [%s] not found '
                         'in url parameter' % (verify_token, challenge))
            return HttpResponse('Error, invalid token')

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        return generic.View.dispatch(self, request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        body = self.request.body.decode('utf-8')
        incoming_message = json.loads(body)
        # logger.debug('DJM_HANDLERS: %s' % DJM_HANDLERS)
        try:
            callback = Callback.deserialize(incoming_message)
            assert isinstance(callback, Callback)
            # save user profile if settings to True
            for entry in callback.entry:
                for messaging in entry.messaging:
                    self._handle_messaging(messaging)
        except Exception as e:
            logging.exception('Got exception on post...')
            logger.error('Failed to handle the message because %s' % e)
        return HttpResponse()

    def _handle_messaging(self, messaging):
        from djmessenger.routing import policy

        logger.debug('*****************************************')
        logger.debug('*****    Start of a new message     *****')
        logger.debug('*****************************************')
        logger.debug(
            'Ready to process messaging of ReceivingType %s: %s' %
            (messaging.get_receiving_type(), messaging.serialize()))
        handlers = policy.get_handlers(messaging.get_receiving_type())
        logger.debug('Got applicable routing handler classes from policy: %s'
                     % handlers)
        # loop through corresponding handlers
        for handler in handlers:
            if not issubclass(handler.get_class(), BaseHandler):
                continue
            handler_instance = handler.get_class()(messaging,
                                                   **handler.get_args())
            handler_name = handler.get_class().__name__
            logger.debug('-->  Start of handling process to handler %s for '
                         'messaging %s <--'
                         % (handler_name, messaging.serialize()))
            if handler_instance.should_handle():
                handler_instance.handle()

                senders = policy.get_senders(messaging.get_receiving_type(),
                                             handler)
                logger.debug('Handler %s is done, ready to send using senders: '
                             '%s' % (handler_name, senders))
                for sender in senders:
                    if not issubclass(sender.get_class(), CommonSender):
                        continue
                    sender_name = sender.get_class().__name__
                    logger.debug(
                        '-->  Start of sending process to sender %s for '
                        'messaging %s <--'
                        % (sender_name, messaging.serialize()))
                    sender_instance = sender.get_class()(handler_instance,
                                                         **sender.get_args())
                    sender_instance.send()
                    logger.debug('-->  End of sending process to %s  <--'
                                 % sender_name)
            else:
                logger.debug('The messaging was not being handled by handler %s'
                             ', so no sending will be performed' %
                             handler_name)
            logger.debug('-->  End of handling process to %s  <--'
                         % handler_name)
        logger.debug('*****************************************')
        logger.debug('*****  End of a processing message  *****')
        logger.debug('*****************************************')



# Copyright 2016 VMware, Inc.  All rights reserved. -- VMware Confidential
# coding: utf-8
from __future__ import absolute_import

try:
    import simplejson as json
except ImportError:
    import json
import codecs

try:
    # python 2
    basestring
except NameError:
    # python 3
    basestring = str

try:
    # python 2
    unicode
except NameError:
    # python 3
    unicode = str

try:
    xrange
except NameError:
    xrange = range

import logging
from collections import OrderedDict

from .decorators import synchronized
from .filelock import SecureOpen


class Serializable(object):
    """
    Serializable provides an implementation to make an object able to be
    serialized into json as in dict or in string format recursively.

    # Notes

    1. When you subclass another Serializable, think about whether you'd like
       to call superclass's constructor. If you did, you need to take care of
       defining includes and excludes for the attributes defined by superclass

    2. includes and excludes can exist at the same time, but you can't define
       the same attribute in both

    3. Always define custom_obj_map, if nothing should be there, put empty dict.
        Otherwise your class might be inheriting it from superclass and you
        don't it which might lead to unexpected result

    4. Always define includes and excludes, if nothing should be there, put
       them as empty tuple, otherwise your class might be inheriting them from
       superclass and you don't know it which might lead to unexpected result

    # Assumptions

    1. All nested custom objects must implement Serializable
    2. If any attribute is custom object and you need to deserialize it later,
       you need to define it in custom_obj_map so that we know the exactly class
        and if it is contained in either list or dict

    Serializable supports list and dict:

    1. Serialization if the attribute is a list: json() will loop through
       the list and construct another list of dict by calling json() for each
       item
    2. Serialization if the attribute is a dict: json() will loop through
       the dict and construct another dict which key remains exactly the same
       but calling value's json()
    3. Deserialization  of custom objects and custom_obj_map was configured for
       attribute name: deserialize() will first identify the attribute type from
        the loaded json dict. If it was a list of custom objects, we de them one
        by one, put them in a list and assign back to the attribute; if it was
        a dict, construct another dict where key remains the same but de the
        value; if it is an object, just de it

    # Configurations

    ## custom_obj_map

        custom_obj_map provides a way to properly deserialize arbitrary json data into
        this class, since this class itself must know what classes its instance
        variables are, this approach makes sense.
        You only need to put the instance variable whose type is not primitive.
        Follow this format:

        ```
        custom_obj_map = {
            'instance variable name': <class>,
        }
        ```

        This way we know exactly the class of the instance variable and whether it
        is a single object or a list of <class>, either way, deserialzation can
        handle it.

        The <class> can be either a class or a string which matches the class
        name, but the class MUST extends Serializable

    ## excludes

        exclude any attributes that are listed in this tuple.

        includes and excludes can exist at the same time but you can't have the
        same attribute name listed in both

    ## includes

        only include attributes that are listed in this tuple

        includes and excludes can exist at the same time but you can't have the
        same attribute name listed in both

    ## exclude_underscore

        exclude any attributes that starts with an underscore

    ## ordering

        If you want to have certain ordering for the class attributes to display
         in the result of serialize(), you can define this list.

        If `ordering` was not defined, the order is not guaranteed.

        If `ordering` is defined, the output of serialization will be ordered
        by it.

        You are not required to include all attributes in `ordering`, but those
        which are defined in `ordering` will be put before those which are not

    # Quick example:

        ```
        class TestDevice(Serializable):
            def __init__(self, id, type):
                # id and type are both primitive, so no need to use _obj_map
                self.id = id
                self.type = type


        class TestUser(Serializable):
            _obj_map = {
                'devices': ['TestDevice', object]
            }
            excludes = ['password']
            exclude_underscore = True

            def __init__(self, username, password, devices=[], log_location=""):
                super(TestUser, self).__init__(log_location)
                self.username = username
                self.password = password
                self.devices = devices

            def add_device(self, device):
                if isinstance(device, TestDevice):
                    self.devices.append(device)
        ```
    """
    custom_obj_map = {}
    includes = ()
    excludes = ()
    exclude_underscode = True
    ordering = []

    class _Empty(object):
        pass

    def serialize(self, pretty=False):
        """
        Serialize the object into a valid json-formatted string after applying
        includes and excludes policies

        @return: json representation for this instance
        @rtype: str
        """
        d = self.json()
        if pretty:
            return json.dumps(d, sort_keys=True, indent=2)
        else:
            return json.dumps(d, sort_keys=False)

    def json(self):
        """
        Returns an OrderedDict representation of the object, after applying
        includes and excludes policies

        This method takes ordering into account

        @return: a dict represents this object recursively
        @rtype: OrderedDict
        """
        if not self.ordering:
            return self.json_unordered()
        ret = OrderedDict()
        # loop through both ordering and __dict__.keys, which means keys in
        # ordering will be processed first, and dup keys will then be ignored
        for k in self.ordering + self.__dict__.keys():
            # already handled k
            if k in ret:
                continue
            if k in self.includes and k in self.excludes:
                raise AssertionError(
                    'You cannot include attribute %s and exclude it at the same'
                    ' time' % k)
            if self.includes and k not in self.includes:
                continue
            if self.excludes and k in self.excludes:
                continue
            if self.exclude_underscode and k.startswith('_'):
                continue
            # get attribute k from self
            v = getattr(self, k, None)
            # somehow k does not exist???
            if v is None:
                continue
            # if value has attribute json, we use it recursively
            if hasattr(v, "json"):
                ret[k] = v.json()
            elif isinstance(v, list):
                ret[k] = [x.json() if hasattr(x, 'json') else x for x in v]
            elif isinstance(v, dict):
                ret[k] = {k1: (v1.json() if hasattr(v1, 'json') else v1) for k1, v1 in v.items()}
            else:
                ret[k] = v
        return ret

    def json_unordered(self):
        """
        This method does not consider ordering

        @return:
        """
        ret = {}
        for k, v in self.__dict__.items():
            if k in self.includes and k in self.excludes:
                raise AssertionError(
                    'You cannot include attribute %s and exclude it at the same'
                    ' time' % k)
            if self.includes and k not in self.includes:
                continue
            if self.excludes and k in self.excludes:
                continue
            # intentionally leave attribute starts with underscore of
            if self.exclude_underscode and k.startswith('_'):
                continue
            if hasattr(v, "json"):
                ret[k] = v.json()
            elif isinstance(v, list):
                ret[k] = [x.json() if hasattr(x, 'json') else x for x in v]
            elif isinstance(v, dict):
                ret[k] = {k1: (v1.json() if hasattr(v1, 'json') else v1) for k1, v1 in v.items()}
            else:
                ret[k] = v
        return ret

    @classmethod
    def deserialize(cls, json_data):
        """
        Provide default implementation of how to deserialize json representation
         to python object.
        If you have any instance variables that are custom objects and you did
        not define them in custom_obj_map, you need to override deserialize()
        otherwise that instance variable will be deserialized to a dict instead
        of the object itself.

        @param json_data: a dict which represents a json object; or a string
                          which is a valid json format
        @type json_data: dict, str

        @return: an instance of this class
        """
        if not json_data:
            raise ValueError('Not able to deserialize %s from data %s' %
                             (cls.__name__, json_data))
        if isinstance(json_data, str):
            json_data = json.loads(json_data)
        elif isinstance(json_data, dict):
            pass
        else:
            # return json_data
            raise ValueError('json_data only takes either string or dict, but '
                             'was given %s' % type(json_data))
        ret = Serializable._Empty()
        for key in json_data:
            obj_map = cls.custom_obj_map
            if key in obj_map:
                # if key in custom_obj_map, we know it is a custom object
                clazz = obj_map[key][0]
                if isinstance(clazz, str):
                    # the class was defined using class name, search subclasses
                    # of Serializable
                    for subclass in Serializable.__subclasses__():
                        if clazz.upper() == subclass.__name__.upper():
                            clazz = subclass
                            break
                    if isinstance(clazz, str):
                        # not found
                        raise ValueError('Given class %s is a string but that '
                                         'class is not found in Serializable '
                                         'subclasses' % clazz)
                elif isinstance(clazz, type):
                    pass
                ttype = obj_map[key][1]
                if ttype == list:
                    # if it is a list of clazz, we need to de them one by one
                    arr = []
                    for sub in json_data[key]:
                        arr.append(clazz.deserialize(sub))
                    ret.__setattr__(key, arr)
                elif ttype == dict:
                    # if it is a dict
                    d = {}
                    for k, v in json_data[key].items():
                        d[k] = clazz.deserialize(v)
                    ret.__setattr__(key, d)
                else:
                    # it if is a single object, just de it
                    obj = clazz.deserialize(json_data[key])
                    ret.__setattr__(key, obj)
            else:
                # if key is not in custom_obj_map, either it is primitive type
                # or definition not provided, we simply put it as a dict
                ret.__setattr__(key, json_data[key])
        ret.__class__ = cls
        return ret

    @synchronized
    def write_json(self, destination):
        """
        Securely opens the log_location and dumps json from self to it, this
        method is synchronized

        @raise ValueError if log_location was not set
        @raise IOError if log_location can not be opened
        """
        with SecureOpen(destination, codecs.open, False, "w", encoding='utf-8') as fp:
            return json.dump(
                self.json(), fp,
                default=lambda o: {k: v for k, v in o.__dict__.items() if
                                   not k.startswith('_')},
                sort_keys=False,
                indent=2)

    @classmethod
    def read_json(cls, json_file):
        """
        Given an absolute file path, try to decode it as json representation
        to an instance of this class

        @param json_file: absolute file path for the .json file
        @type json_file: str

        @return: an instance of this class
        @raise ValueError: if the json representation can not be deserialized
        @raise IOError: if the file can not be opened
        @raise ValueError: if not able to deserialize the json data into cls
        """
        with open(json_file) as fp:
            json_data = json.load(fp)
            try:
                ret = cls.deserialize(json_data)
            except ValueError as err:
                raise ValueError("The json file provided was not "
                                 "serialized from this class. Underlying error"
                                 "message is %s" % (err.message))
            return ret

    def __eq__(self, other):
        """
        Override __eq__ magic method to compare to Serializables recursively

        @param other:
        @type other: Serializable

        @return: True if the 2 objects are recursively equal; False otherwise
        @rtype: bool
        """
        if not isinstance(other, type(self)):
            return False
        for name, obj in vars(self).items():
            # the following 3 conditions are to honor the include/exclude
            # policy, since either part of the __eq__ method could be an object
            # deserialized from dict or string, the resulting object might not
            # have some attributes
            if self.includes and name not in self.includes:
                continue
            if self.excludes and name in self.excludes:
                continue
            # don't compare attributes startswith underscore
            if self.exclude_underscode and name.startswith('_'):
                continue
            # other does not have this attribute
            if name not in vars(other):
                return False
            # python 2: str and unicode are different classes and deserialized
            # string is unicode, so check with basestring and convert it to
            # unicode for comparison
            if isinstance(obj, basestring):
                obj = unicode(obj)
            other_obj = other.__getattribute__(name)
            if isinstance(other_obj, basestring):
                other_obj = unicode(other_obj)
            # if the attribute was a list before serialization, it will become
            # a tuple after deserialization, so converting both to tuple for
            # comparison
            if isinstance(obj, list):
                obj = tuple(obj)
            if isinstance(other_obj, list):
                other_obj = tuple(other_obj)
            # the attribute types are different
            if not isinstance(other_obj, type(obj)):
                return False
            # only need to check with tuple since we converted all lists to
            # tuples
            if issubclass(obj.__class__, tuple):
                if len(obj) != len(other_obj):
                    return False
                for index in xrange(len(obj)):
                    item = obj[index]
                    other_item = other_obj[index]
                    if hasattr(item, '__eq__') and not item.__eq__(other_item):
                        return False
                    elif not hasattr(item, '__eq__') and item != other_item:
                        return False
            elif issubclass(obj.__class__, dict):
                if len(obj) != len(other_obj):
                    return False
                for key in obj.keys():
                    item = obj[key]
                    other_item = other_obj[key]
                    if hasattr(item, '__eq__') and not item.__eq__(other_item):
                        return False
                    elif not hasattr(item, '__eq__') and item != other_item:
                        return False
            elif hasattr(obj, '__eq__') and not obj.__eq__(other_obj):
                return False
            elif not hasattr(obj, '__eq__') and obj != other_obj:
                return False
        return True


class SerializableEnum(Serializable):
    """
    An Enum-like class, which serializes to a simple string. In json
    serialization, this will be treated just like a string, instead of an
    object

    Usage:

    >>> class Severity(SerializableEnum):
    ...    pass
    >>> Severity.INFO = Severity('INFO', 'INFO')
    >>> Severity.WARNING = Severity('WARNING', 'WARNING')
    >>> Severity.ERROR = Severity('ERROR', 'ERROR')
    >>> Severity.INFO.name == 'INFO'
    True
    >>> Severity.value_of('info') == Severity.INFO
    True
    >>> Severity.value_of('ERROR') == Severity.ERROR
    True
    >>> Severity.value_of('ERROR') == Severity.WARNING
    False
    >>> test = Severity.value_of('TEST')
    Traceback (most recent call last):
    ...
    KeyError: 'name TEST not defined for Severity'
    >>> Severity.INFO.serialize() == 'INFO'
    True
    """
    # _members = set()

    def __init__(self, name, description=""):
        """

        @param name: unique identifier, case sensitive
        @type name: str

        @param description: messages to display
        @type description: str
        """
        self.name = name
        self._description = description
        # self._members.add(self)
        if hasattr(self.__class__, 'members'):
            self.__class__.members[self.name] = self
        else:
            s = dict()
            s[self.name] = self
            setattr(self.__class__, 'members', s)

    def __hash__(self):
        return hash(('%s.%s' % (self.__module__, self.__class__.__name__), self.name))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.name == other.name

    def serialize(self, pretty=False):
        return self.name

    @classmethod
    def deserialize(cls, json_data):
        """

        @param json_data: this must be a single string
        @return: An instance of Status
        @rtype: Status
        """
        return cls.value_of(json_data)

    def json(self):
        return self.name

    def __repr__(self):
        return '(%s) %s' % (self.__class__.__name__, self.name)

    def __str__(self):
        return self.name

    @classmethod
    def get_members(cls):
        """

        @return: A map maps from enum name to the actual enum instance
        """
        # return {k for k in cls._members if isinstance(k, cls)}
        return getattr(cls, 'members')

    @classmethod
    def value_of(cls, name):
        """
        Given enum name, find and return the corresponding enum object

        @param name: A string for the status code
        @param name: str

        @return: An instance of Status
        @rtype: Status
        @raise KeyError: if name is not defined for the enum
        """
        assert isinstance(name, basestring)
        # for obj in cls.members():
        #     if obj.name == name:
        #         return obj
        if name in cls.get_members():
            return cls.get_members()[name]
        raise KeyError('name %s not defined for %s' % (name, cls.__name__))


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
import requests
import gettext
import logging

from django.core.management.base import BaseCommand

from djmessenger.settings import DJM_THREAD_SETTINGS_URL, DJM_THREAD_PERSISTENT_MENUS
from djmessenger.payload import *


logger = logging.getLogger(__name__)


class Command(BaseCommand):
    def handle(self, *args, **options):
        logger.debug('Ready to configure Persistent Menu')
        if not DJM_THREAD_PERSISTENT_MENUS:
            return
        actions = []
        # let index starts from 1
        for index in range(1, len(DJM_THREAD_PERSISTENT_MENUS) + 1):
            item = DJM_THREAD_PERSISTENT_MENUS[index-1]
            if index > 5:
                logger.debug('Facebook only permits 5 Persistent Menu items, '
                             'this item is the %d item, skipping' % index)
                break
            if 'title' not in item:
                logger.debug('There is no title in item, skipping')
                continue
            if 'url' in item:
                ttype = 'web_url'
                actions.append({
                    "type": ttype,
                    "title": item['title'],
                    "url": item['url']
                })
            else:
                ttype = 'postback'
                if index == 1:
                    payload = PersistentMenuOnePayload().serialize()
                elif index == 2:
                    payload = PersistentMenuTwoPayload().serialize()
                elif index == 3:
                    payload = PersistentMenuThreePayload().serialize()
                elif index == 4:
                    payload = PersistentMenuFourPayload().serialize()
                elif index == 5:
                    payload = PersistentMenuFivePayload().serialize()
                actions.append({
                    "type": ttype,
                    "title": item['title'],
                    "payload": payload
                })

        data = {
            "setting_type": "call_to_actions",
            "thread_state": "existing_thread",
            "call_to_actions": actions
        }
        logger.debug('Prepared payload for configuring Persistent Menu as %s'
                     % data)
        status = requests.post(DJM_THREAD_SETTINGS_URL,
                               headers={"Content-Type": "application/json"},
                               json=data)
        print(status.content)
        logger.debug('Configuring Persistent Menu got response code %s, '
                     'content %s' % (status.status_code, status.content))



# -*- coding: utf-8 -*-
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
import os
from django.utils.translation import ugettext as _
from djmessenger.utils.default_routing_policy import DJM_DEFAULT_ROUTING_POLICY


# You don't need to change this
DJM_BASE_DIR = os.path.dirname(os.path.abspath(__file__))

# Facebook Page Access Token, get it from developers.facebook.com
DJM_PAGE_ACCESS_TOKEN = getattr(settings, 'DJM_PAGE_ACCESS_TOKEN', '')

# The endpoint that Facebook will relay the callback message to
# also being used to setup webhook
DJM_ENDPOINT = getattr(settings, 'DJM_ENDPOINT', '')

# You don't need to change this
DJM_POST_MESSAGE_URL = getattr(settings, 'DJM_POST_MESSAGE_URL',
                               'https://graph.facebook.com/v2.6/me/messages?access_token=%s' % DJM_PAGE_ACCESS_TOKEN)

# You don't need to change this
DJM_USER_DETAILS_URL = getattr(settings, 'DJM_USER_DETAILS_URL',
                               'https://graph.facebook.com/v2.6/%s')

# You don't need to change this
DJM_THREAD_SETTINGS_URL = getattr(settings, 'DJM_THREAD_SETTINGS_URL',
                                  'https://graph.facebook.com/v2.6/me/thread_settings?access_token=%s' % DJM_PAGE_ACCESS_TOKEN)

# Whether DJM should automatically fetch and save user profile for any user that
# sends message to the page and observed by BOT, default to True
DJM_SAVE_USER_PROFILE = getattr(settings, 'DJM_SAVE_USER_PROFILE', True)


# A sender named djmessenger.sending.DefaultSender which is a SimpleTextSender
# you can define your message here and directly use it in DJM_ROUTING_POLICY
DJM_DEFAULT_SENDER_TEXT = getattr(settings, 'DJM_DEFAULT_SENDER_TEXT',
                                  _('Thanks for your message'))

# Prefix this string to all text messages sent to users
DJM_BOT_PREFIX = getattr(settings, 'DJM_BOT_PREFIX', _("BOT: "))

"""
## DJM_ROUTING_POLICY

DJM_ROUTING_POLICY defines the lifecycle of a message being received, handling
it and lastly send some response back. In the following format, each router
represents a type of message received, handlers represent which handlers should
be used to handle this type of message and lastly which senders should be used
to send the response back.

DEFAULT is a special type, it is not a type that Facebok would send over but it
is merely for DJM to know that the handlers listed in DEFAULT should apply to
all other types as well.

For example, in the following default policy, router[1], it basically means

1. If we got a message type is sticker
2. send the message to ThumbUpHandler
3. ThumbUpHandler.should_handle() will check whether it is valid to handle this
   message, if yes, handle it
4. Then senders defines which Sender class should be used to send something back
"""
DJM_ROUTING_POLICY = getattr(settings, 'DJM_ROUTING_POLICY',
                             DJM_DEFAULT_ROUTING_POLICY)
"""
https://developers.facebook.com/docs/messenger-platform/thread-settings/greeting-text

1. length must be <= 160
2. only first time visitor sees this greeting text
"""
DJM_THREAD_GREETINGS_TEXT = getattr(settings, 'DJM_THREAD_GREETINGS_TEXT',
                                    _("Welcome to the page!"))
"""
https://developers.facebook.com/docs/messenger-platform/thread-settings/persistent-menu

1. Only allow at most 5 menu items
2. Items without url will send a Postback Message to the server just like a Button
3. Items with url will simply open that url on a browser
4. title has 30 char limit
5. The order of items will be displayed to the user in the same order here
"""
DJM_THREAD_PERSISTENT_MENUS = getattr(settings, 'DJM_THREAD_PERSISTENT_MENUS',
                                      [
                                          {
                                              "title": "Help",
                                          },
                                          {
                                              "title": "Visit Website",
                                              "url": ""
                                          }
                                      ])
"""
https://developers.facebook.com/docs/messenger-platform/thread-settings/get-started-button
Whether to enable get started button feature
"""
DJM_THREAD_ENABLE_GET_STARTED_BUTTON = getattr(
    settings, 'DJM_THREAD_ENABLE_GET_STARTED_BUTTON', False)
"""
If set to True, djmessenger will send POST request to configure thread
automatically. If set to False, you need to configure it by using management
commands
"""
# TODO: not implemented yet
DJM_THREAD_ENABLE_AUTO_CONFIG = getattr(
    settings, 'DJM_THREAD_ENABLE_AUTO_CONFIG', False)

# prechecks
if not DJM_PAGE_ACCESS_TOKEN or not DJM_ENDPOINT:
    raise ImproperlyConfigured(
        _('djmessenger requires at least DJM_PAGE_ACCESS_TOKEN and DJM_ENDPOINT'
          ' to be configured in your settings.py')
    )



# -*- coding: utf-8 -*-
from ..models import FBUserProfile
import gettext
import logging
from django.conf import settings


logger = logging.getLogger(__name__)


def install_user_locale(psid):
    """
    Install the locale for the user psid, if not able to load it, reset

    @param psid:
    @return:
    """
    try:
        user = FBUserProfile.objects.get(pk=psid)
        lang = user.locale
        logger.debug('locale dirs: %s' % settings.LOCALE_PATHS)
        for locale_dir in settings.LOCALE_PATHS:
            zh = gettext.translation('django',
                                     localedir=locale_dir,
                                     languages=[lang])
            zh.install()
            logger.debug('Successfully installed language %s for user %s'
                         % (lang, psid))
    except Exception as e:
        logger.debug('Installing locale for user %s failed because %s, '
                     'resetting...' % (psid, e))
        reset_locale()


def reset_locale():
    logger.debug('Resetting locale...')
    _ = lambda s: s



# -*- coding: utf-8 -*-
"""
Similarly, this sending module is to represent the json templates that should
be used when sending something back to the user based on Facebook reference:

https://developers.facebook.com/docs/messenger-platform/send-api-reference

## CommonSender

`CommonSender` defines 4 attributes: `recipient`, `sender_action`,
`notification_type`, `message` because that's what [Facebook Send API](https://developers.facebook.com/docs/messenger-platform/send-api-reference) requires

And `send()` method actually sends the json string (serialized from the above
4 attributes) to Facebook endpoint using `requests` module

Each sender will definitely get its handler as its first constructor argument.
You can define your own sender to provide customized behavior and then use it
in the routing policy

## Provided Senders

### ButtonSender (TBD)

### DefaultSender

Simple sends back the text defined in `DJM_DEFAULT_SENDER_TEXT`

- args: takes no argument

### MultimediaSender

Sends back a `SendingType` with its url

    ```
    image
    audio
    video
    file
    ```

- args:
    - sending_type: image, audio, video, file
    - url: url to the resource

### QuickReplySender (TBD)

### SenderActionSender

Sends back a sender action, check [Facebook Reference](https://developers.facebook.com/docs/messenger-platform/send-api-reference/sender-actions)

- args:
    - action: mark_seen, typing_on, typing_off

### SimpleMessageSender

Sends back a simple text message, could be i18n'ed

- args:
    - text: message to send back to user

"""
from abc import abstractmethod

import requests
import gettext

from djmessenger.exceptions import DJMInvalidConfigException
from djmessenger.models import Sending
from djmessenger.payload import Payload
from djmessenger.receiving import Recipient
from djmessenger.settings import DJM_POST_MESSAGE_URL, DJM_DEFAULT_SENDER_TEXT, DJM_BOT_PREFIX
from djmessenger.utils.i18n import *
from djmessenger.utils.serializable import SerializableEnum, Serializable
from djmessenger.utils.utils import get_class_name
from djmessenger.handling import BaseHandler


logger = logging.getLogger(__name__)


class SendingType(SerializableEnum):
    pass


SendingType.SIMPLE_TEXT = SendingType('simple_text')
SendingType.QUICK_REPLY = SendingType('quick_reply')
SendingType.BUTTON = SendingType('button')
SendingType.IMAGE = SendingType('image')
SendingType.AUDIO = SendingType('audio')
SendingType.VIDEO = SendingType('video')
SendingType.FILE = SendingType('file')
SendingType.SENDER_ACTION = SendingType('sender_action')


class NotificationType(SerializableEnum):
    pass


NotificationType.REGULAR = NotificationType('REGULAR')
NotificationType.SILEN_PUSH = NotificationType('SILEN_PUSH')
NotificationType.NO_PUSH = NotificationType('NO_PUSH')


class SenderAction(SerializableEnum):
    pass


SenderAction.MARK_SEEN = SenderAction('mark_seen')
SenderAction.TYPING_ON = SenderAction('typing_on')
SenderAction.TYPING_OFF = SenderAction('typing_off')


class ButtonType(SerializableEnum):
    pass


ButtonType.WEB_URL= ButtonType('web_url')
ButtonType.POSTBACK = ButtonType('postback')


class CommonSender(Serializable):
    """
    Subclasses of CommonSender must remember to call super().__init__() and the
    first argument must be a handler
    """
    # excludes = ('handler',)

    def __init__(self, handler, sender_action=None,
                 notification_type=NotificationType.REGULAR.name):
        """
        Define highest level of sending template

        @param sender_action: sender action
        @type sender_action: SenderAction

        @param notification_type: mostly no need to change this, default is
                                  regular
        @type notification_type: NotificationType
        """
        assert issubclass(type(handler), BaseHandler), \
            'The first argument in CommonSender constructor must be a subclass ' \
            'of BaseHandler'
        self.recipient = Recipient(handler.get_psid())
        self.sender_action = sender_action
        self.notification_type = notification_type
        self.message = None
        self._handler = handler

    @abstractmethod
    def get_message(self):
        """
        Based on the attributes the class constructor constructs, this method
        returns the message body, the returned must be a dict and valid json
        object.

        When construct the request json payload, if there is any text that
         could be i18n'ed, please make sure to use

        `CommonSender.preprocess_outgoing_string()` to make sure the string is
        using correct locale

        @return:
        @rtype: dict
        """
        pass

    def send(self):
        self.message = self.get_message()
        data = self.json()
        logger.debug('Sending %s message to user %s: %s' %
                     (self.get_sending_type(), self.recipient.id, data))
        status = requests.post(DJM_POST_MESSAGE_URL,
                               headers={"Content-Type": "application/json"},
                               json=data)
        try:
            recipient = FBUserProfile.objects.get(pk=self.recipient.id)
            try:
                sending = Sending.objects.create(recipient=recipient, data=data,
                                                 type=self.get_sending_type().name)
                sending.response = status.content
                sending.status_code = status.status_code
                if status.status_code == 200:
                    logger.debug('Successfully sent message %s' % data)
                else:
                    logger.debug('Sent message %s failed with status code %d'
                                 'and failure details %s' %
                                 (data, status.status_code, status.content))
                sending.save()
            except Exception as e:
                logger.debug('Failed to create Sending object because %s' % e)
        except FBUserProfile.DoesNotExist:
            logger.debug('Recipient PSID not found in database table djm_user,'
                         ' since PSID was not found, not able to save what are '
                         'we sending')

    @abstractmethod
    def get_sending_type(self):
        """
        Return the SendingType that this Sender sends

        @return:
        @rtype: SendingType
        """
        pass

    @classmethod
    def preprocess_outgoing_string(cls, psid, text, add_prefix=True):
        """
        This method take cares of any preprocessing of the outgoing text.
        More specifically, this method converts the text into the user's locale,
         if available.

        @param psid:
        @param text:
        @param add_prefix: whether prepend DJM_BOT_PREFIX to the start of text,
                           you might want to use it when processing button text
        @type add_prefix: bool
        @return:
        """
        install_user_locale(psid)
        try:
            if add_prefix:
                ret = _(DJM_BOT_PREFIX).encode('utf-8') + _(text).encode('utf-8')
            else:
                ret = _(text).encode('utf-8')
        except:
            # if the locale is not supported, _() will fail, so catch it and
            # fallback to English
            if add_prefix:
                ret = DJM_BOT_PREFIX.encode('utf-8') + text.encode('utf-8')
            else:
                ret = text.encode('utf-8')
        reset_locale()
        return ret

    def get_psid(self):
        return self.recipient.id


class SenderActionSender(CommonSender):
    """
    Send a Sender Action
    """
    def __init__(self, handler, action):
        super().__init__(handler, action)

    def get_sending_type(self):
        return SendingType.SENDER_ACTION

    def get_message(self):
        # no message
        return ""


class SimpleMessageSender(CommonSender):
    """
    Send a simple text message
    """
    excludes = ['text']

    def __init__(self, handler, text):
        super().__init__(handler)
        self.text = text

    def get_message(self):
        # FUTURE: we can probably move this part to CommonSender so that
        #         subclasses don't need to worry about it
        text = CommonSender.preprocess_outgoing_string(
            self.get_psid(), self.text)
        if isinstance(text, bytes):
            text = text.decode('utf-8')
        return {"text": text}

    def get_sending_type(self):
        return SendingType.SIMPLE_TEXT


class DefaultSender(SimpleMessageSender):
    """
    Sends back simple text DJM_DEFAULT_SENDER_TEXT
    """
    def __init__(self, handler):
        super().__init__(handler, DJM_DEFAULT_SENDER_TEXT)


class MultimediaSender(CommonSender):
    # TODO: should work with django static???
    """
    Send back multimedia from a url

    SendingType.IMAGE = SendingType('IMAGE')
    SendingType.AUDIO = SendingType('AUDIO')
    SendingType.VIDEO = SendingType('VIDEO')
    SendingType.FILE = SendingType('FILE')
    """

    def __init__(self, handler, sending_type, url):
        super().__init__(handler)
        self.sending_type = sending_type
        self.url = url

    def get_sending_type(self):
        return SendingType.value_of(self.sending_type)

    def get_message(self):
        return {"attachment": {
            "type": self.get_sending_type().name.lower(),
            "payload": {
                "url": self.url
            }
        }}


class BaseButtonSender(SimpleMessageSender):
    """
    Facebook restricts to have 3 buttons at a time
    """
    LIMIT = 3

    class Button(Serializable):
        def __init__(self, title, url="", payload=None):
            """
            There are 2 types of buttons: web_url and postback where web_url
            button simply open the url in a browser and postback sends a
            Postback message back to the BOT

            @param button_type: web_url or postback
            @type button_type: ButtonType

            @param title: button title
            @type title: str

            @param url: only applicable if button_type == web_url
            @type url: str

            @param payload: only applicable if button_type == postback
            @type payload: subclass of Payload.serialize()
            """
            if not url and not payload:
                raise ValueError('url and payload can not both be empty')
            if url and payload:
                raise ValueError('url and payload are mutually-exclusive')
            self.type = ButtonType.WEB_URL if url else ButtonType.POSTBACK
            self.title = title
            if url:
                self.url = url
            if payload and not issubclass(type(payload), Payload):
                raise ValueError('Payload for Button must be a subclass of'
                                 '%s' % get_class_name(Payload))
            if payload:
                self.payload = payload.serialize()

    def __init__(self, handler, text):
        """
        text if the text that appears in the main body

        @param psid:
        @param text:
        @type text: str
        """
        super().__init__(handler, text)
        self.buttons = list()
        self.template_type = "button"

    def add_button(self, title, url="", payload=None):
        if len(self.buttons) == BaseButtonSender.LIMIT:
            raise DJMInvalidConfigException('Buttons can have only 3 at a time')
        button = BaseButtonSender.Button(
            CommonSender.preprocess_outgoing_string(self.get_psid(), title, False),
            url, payload)
        self.buttons.append(button)

    def get_message(self):
        bts = []
        for bt in self.buttons:
            bts.append(bt.json())
        ret = {
            "attachment": {
                "type": "template",
                "payload": {
                    "template_type": "button",
                    "text": CommonSender.preprocess_outgoing_string(self.get_psid(), self.text),
                    "buttons": bts
                }
            }
        }
        return ret

    def get_sending_type(self):
        return SendingType.BUTTON


class BaseQuickReplySender(SimpleMessageSender):
    """
    Facebook restricts to have 10 quick replies at a time
    """
    LIMIT = 10

    class QuickReply(Serializable):
        def __init__(self, title="", payload=None):
            """

            @param title:
            @param payload:
            @type payload: subclass of Payload.serialize()
            """
            self.title = title
            if not issubclass(type(payload), Payload):
                raise ValueError('Payload for QuickReply must be a subclass of'
                                 '%s' % get_class_name(Payload))
            self.payload = payload.serialize()
            self.content_type = "text"

    def __init__(self, handler, text):
        """
        Quick Reply is essentially a SimpleMessageSender with buttons, then you
        send a quick reply, the user will first see the text just like what they
         will from a SimpleMessageSender, follow by a list of buttons they can
         click on. Upon clicking, it is sending another SIMPLE_TEXT back to the
         server with quick_reply component

        @param psid:
        @param text:
        """
        super().__init__(handler, text)
        self.quick_replies = []

    def add_quick_reply(self, title, payload):
        """
        An entry of quick reply button is represented like this

        ```
        {
            "content_type":"text",
            "title":"Red",
            "payload":"DEVELOPER_DEFINED_PAYLOAD_FOR_PICKING_RED"
        },
        ```
        content_type must be "text", so the arguments here are title and payload
        , payload must be a subclass of Payload

        @return:
        """
        if len(self.quick_replies) == BaseQuickReplySender.LIMIT:
            raise DJMInvalidConfigException('Quick Reply can only have 10')
        qr = BaseQuickReplySender.QuickReply(title, payload)
        self.quick_replies.append(qr)

    def get_message(self):
        data = {
            "text": CommonSender.preprocess_outgoing_string(self.get_psid(),
                                                            self.text)
        }
        qrs = []
        for qr in self.quick_replies:
            qrs.append(qr.json())
        data['quick_replies'] = qrs
        return data

    def get_sending_type(self):
        return SendingType.QUICK_REPLY



# -*- coding: utf-8 -*-
"""
This management command will perform makemessages for all locales defined by
Facebook
"""
from django.core.management.base import BaseCommand
from django.core.management.commands import makemessages
import xmltodict
import os


class Command(BaseCommand):
    def handle(self, *args, **options):
        file = os.path.join(os.path.dirname(__file__),
                            '../../../FacebookLocales.xml')
        if not os.path.exists(file):
            raise IOError('FacebookLocales.xml not found in %s' % file)
        locales = []
        with open(file) as fd:
            doc = xmltodict.parse(fd.read())
            for locale in doc['locales']['locale']:
                representation = locale['codes']['code']['standard']['representation']
                locales.append(representation)
        makemessages.Command().handle(
            **{
                'locale': locales,
                'exclude': [],
                'domain': 'django',
                'all': False,
                'extensions': [],
                'symlinks': False,
                'ignore': [],
                'no-default-ignore': True,
                'no_wrap': False,
                'no_location': False,
                'no_obsolete': False,
                'keep_pot': False,
                'verbosity': 1,
                'ignore_patterns': [],
                'use_default_ignore_patterns': True
            }
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2016-08-24 06:25
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('djmessenger', '0008_auto_20160824_0519'),
    ]

    operations = [
        migrations.AddField(
            model_name='fbuserprofile',
            name='thumbups',
            field=models.BigIntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='fbuserprofile',
            name='first_name',
            field=models.CharField(blank=True, max_length=512, null=True, verbose_name='first name'),
        ),
        migrations.AlterField(
            model_name='fbuserprofile',
            name='gender',
            field=models.CharField(blank=True, max_length=20, null=True),
        ),
        migrations.AlterField(
            model_name='fbuserprofile',
            name='is_active',
            field=models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name='fbuserprofile',
            name='last_name',
            field=models.CharField(blank=True, max_length=512, null=True, verbose_name='last name'),
        ),
        migrations.AlterField(
            model_name='fbuserprofile',
            name='locale',
            field=models.CharField(blank=True, max_length=128, null=True),
        ),
        migrations.AlterField(
            model_name='fbuserprofile',
            name='profile_pic',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='fbuserprofile',
            name='psid',
            field=models.CharField(max_length=512, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='fbuserprofile',
            name='timezone',
            field=models.SmallIntegerField(),
        ),
        migrations.AlterField(
            model_name='messaging',
            name='body',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='messaging',
            name='type',
            field=models.CharField(blank=True, max_length=128, null=True),
        ),
        migrations.AlterField(
            model_name='persistentmenu',
            name='type',
            field=models.CharField(choices=[('postback', 'postback'), ('web_url', 'web_url')], default='web_url', help_text='postback means when clicking on the menu, it will send the payload back to the server; while web_url simply opens the url', max_length=10),
        ),
        migrations.AlterField(
            model_name='userlocation',
            name='latitude',
            field=models.DecimalField(blank=True, decimal_places=8, max_digits=10, null=True),
        ),
        migrations.AlterField(
            model_name='userlocation',
            name='longitude',
            field=models.DecimalField(blank=True, decimal_places=8, max_digits=11, null=True),
        ),
        migrations.AlterField(
            model_name='userlocation',
            name='mid',
            field=models.CharField(blank=True, max_length=512, null=True),
        ),
        migrations.AlterField(
            model_name='userlocation',
            name='seq',
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='userlocation',
            name='timestamp',
            field=models.BigIntegerField(help_text='Facebook returns timestamp as EPOCH'),
        ),
        migrations.AlterField(
            model_name='userlocation',
            name='url',
            field=models.URLField(blank=True, max_length=1000, null=True),
        ),
        migrations.AlterField(
            model_name='userlocation',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='djmessenger.FBUserProfile'),
        ),
    ]



# -*- coding: utf-8 -*-
import requests
import logging

from django.db import models
from django.utils.translation import ugettext_lazy as _

from djmessenger.utils.serializable import Serializable
from djmessenger.exceptions import DJMInvalidConfigException


logger = logging.getLogger(__name__)


class Messaging(models.Model):
    body = models.TextField()
    type = models.CharField(max_length=128, null=True, blank=True)

    class Meta:
        db_table = 'djm_messaging'


class Sending(models.Model):
    recipient = models.ForeignKey('FBUserProfile')
    date_sent = models.DateTimeField(auto_now_add=True)
    data = models.TextField()
    type = models.CharField(max_length=128, null=True, blank=True)
    response = models.TextField(null=True, blank=True)
    status_code = models.IntegerField(null=True, blank=True)

    class Meta:
        db_table = 'djm_sending'


class FBUserProfile(models.Model):
    """
    This FacebookUser class is to represent a user that sent a message to us
    via Facebook Messenger, from the message relayed by Facebook, we can look
    up user details which will be stored here
    """
    psid = models.CharField(max_length=512, primary_key=True)
    first_name = models.CharField('first name', max_length=512, null=True,
                                  blank=True)
    last_name = models.CharField('last name', max_length=512, null=True,
                                 blank=True)
    profile_pic = models.TextField(null=True, blank=True)
    locale = models.CharField(max_length=128, null=True, blank=True)
    timezone = models.SmallIntegerField()
    gender = models.CharField(max_length=20, null=True, blank=True)
    is_active = models.BooleanField(default=True)
    date_joined = models.DateTimeField(auto_now_add=True)
    thumbups = models.BigIntegerField(default=0)

    class Meta:
        db_table = 'djm_user'


class UserLocation(models.Model):
    """
    From the message user sent via Facebook to us, addition to saving the raw
    request body into RawBody, we try to see if the sent message is actually a
    location that contains lat and long, if yes, then we save it here
    """
    user = models.ForeignKey(FBUserProfile, null=False, blank=False)
    latitude = models.DecimalField(max_digits=10,
                                   decimal_places=8,
                                   null=True, blank=True)
    longitude = models.DecimalField(max_digits=11,
                                    decimal_places=8,
                                    null=True, blank=True)
    timestamp = models.BigIntegerField(
        help_text='Facebook returns timestamp as EPOCH')
    url = models.URLField(max_length=1000, null=True, blank=True)
    mid = models.CharField(max_length=512, null=True, blank=True)
    seq = models.IntegerField(blank=True, null=True)
    date_created = models.DateTimeField(auto_now_add=True)

    # def save(self, *args, **kwargs):
    #     self.date_created = datetime.utcnow().replace(tzinfo=utc)
    #     super().save(*args, **kwargs)

    class Meta:
        get_latest_by = 'date_created'
        db_table = 'djm_userlocation'

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return '%s(%s, %s)' % (self.user, self.latitude, self.longitude)



# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2016-08-23 23:37
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djmessenger', '0005_auto_20160823_2249'),
    ]

    operations = [
        migrations.CreateModel(
            name='Messaging',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('body', models.TextField()),
                ('type', models.CharField(blank=True, max_length=128, null=True)),
            ],
        ),
        migrations.DeleteModel(
            name='RawBody',
        ),
    ]



# -*- coding: utf-8 -*-
"""
# Receiving module

This module contains all classes that represent a Facebook callback.
Basically this is just models to represent the json templates that Facebook
would sent over as described here:

[Facebook Webhook Reference](https://developers.facebook.com/docs/messenger-platform/webhook-reference)

"""
import time
from abc import ABC, abstractmethod
from decimal import Decimal
import logging
import json

from djmessenger.utils.serializable import Serializable, SerializableEnum
from djmessenger.payload import *


logger = logging.getLogger(__name__)


class ReceivingType(SerializableEnum):
    """
    ReceivingType defines all types that will be sent by Facebook
    """
    pass

ReceivingType.SIMPLE_TEXT = ReceivingType('SIMPLE_TEXT')
ReceivingType.QUICK_REPLY = ReceivingType('QUICK_REPLY')
ReceivingType.IMAGE = ReceivingType('IMAGE')
ReceivingType.AUDIO = ReceivingType('AUDIO')
ReceivingType.VIDEO = ReceivingType('VIDEO')
ReceivingType.FILE = ReceivingType('FILE')
ReceivingType.LOCATION = ReceivingType('LOCATION')
ReceivingType.STICKER = ReceivingType('STICKER')
ReceivingType.POSTBACK = ReceivingType('POSTBACK')
ReceivingType.AUTHENTICATION = ReceivingType('AUTHENTICATION')
ReceivingType.ACCOUNT_LINKING = ReceivingType('ACCOUNT_LINKING')
ReceivingType.DELIVERED = ReceivingType('DELIVERED')
ReceivingType.READ = ReceivingType('READ')
ReceivingType.ECHO = ReceivingType('ECHO')
# this is not a real type, this is for routing.py to indicate the settings is
# not for a specific type but for all type
ReceivingType.DEFAULT = ReceivingType('DEFAULT')
# the follows are special types to handle thread settings, all of these will
# be handled by PostbackReceivedChecker
ReceivingType.PERSISTENT_MENU_ONE = ReceivingType('PERSISTENT_MENU_ONE')
ReceivingType.PERSISTENT_MENU_TWO = ReceivingType('PERSISTENT_MENU_TWO')
ReceivingType.PERSISTENT_MENU_THREE = ReceivingType('PERSISTENT_MENU_THREE')
ReceivingType.PERSISTENT_MENU_FOUR = ReceivingType('PERSISTENT_MENU_FOUR')
ReceivingType.PERSISTENT_MENU_FIVE = ReceivingType('PERSISTENT_MENU_FIVE')
ReceivingType.GET_STARTED_BUTTON = ReceivingType('GET_STARTED_BUTTON')


class MessagingChecker(ABC):
    """
    MessagingChecker is to check a messaging that ReceivingType it should be
    """
    def __init__(self, messaging):
        """

        @param messaging:
        @type messaging: Messaging
        """
        self.messaging = messaging

    @abstractmethod
    def check(self):
        """

        @return: Corresponding CallbackType if it passed any check; None if
                 nothing matches
        @rtype: ReceivingType
        """
        pass


class MessageReceivedChecker(MessagingChecker):
    """
    This check handles all types in MessagedReceived and MessageEcho
    """
    def check(self):
        if not hasattr(self.messaging, 'message'):
            return None
        # first check quick reply because it also has text
        if self.messaging.message is not None \
                and hasattr(self.messaging.message, 'quick_reply') \
                and self.messaging.message.quick_reply is not None \
                and self.messaging.message.quick_reply.payload:
            return ReceivingType.QUICK_REPLY
        # check sticker
        if self.messaging.message is not None \
                and hasattr(self.messaging.message, 'sticker_id') \
                and self.messaging.message.sticker_id is not None:
            return ReceivingType.STICKER
        # then check attachments
        if self.messaging.message is not None \
                and hasattr(self.messaging.message, 'attachments') \
                and self.messaging.message.attachments:
            attachments = self.messaging.message.attachments
            if len(attachments) > 0:
                attachment = attachments[0]
                atype = attachment.type
                try:
                    obj = AttachmentType.value_of(atype)
                except KeyError:
                    raise KeyError('AttachmentType invalid %s' % atype)
                if obj == AttachmentType.IMAGE:
                    return ReceivingType.IMAGE
                elif obj == AttachmentType.AUDIO:
                    return ReceivingType.AUDIO
                elif obj == AttachmentType.VIDEO:
                    return ReceivingType.VIDEO
                elif obj == AttachmentType.FILE:
                    return ReceivingType.FILE
                elif obj == AttachmentType.LOCATION:
                    return ReceivingType.LOCATION
        # check echo
        if self.messaging.message is not None \
                and hasattr(self.messaging.message, 'is_echo') \
                and self.messaging.message.is_echo is not None \
                and self.messaging.message.is_echo:
            return ReceivingType.ECHO
        # if failed, then check whether it has text in message
        if self.messaging.message is not None \
                and hasattr(self.messaging.message, 'text') \
                and self.messaging.message.text:
            return ReceivingType.SIMPLE_TEXT


class PostbackReceivedChecker(MessagingChecker):
    def check(self):
        if not hasattr(self.messaging, 'postback'):
            return None
        if self.messaging.postback is not None \
                and self.messaging.postback.payload:
            payload_string = self.messaging.postback.payload
            logger.debug('Got postback payload string: [%s], determining more '
                         'granular types' % payload_string)
            try:
                subpayload = Payload.get_instance(payload_string)
                if isinstance(subpayload, PersistentMenuOnePayload):
                    return ReceivingType.PERSISTENT_MENU_ONE
                elif isinstance(subpayload, PersistentMenuTwoPayload):
                    return ReceivingType.PERSISTENT_MENU_TWO
                elif isinstance(subpayload, PersistentMenuThreePayload):
                    return ReceivingType.PERSISTENT_MENU_THREE
                elif isinstance(subpayload, PersistentMenuFourPayload):
                    return ReceivingType.PERSISTENT_MENU_FOUR
                elif isinstance(subpayload, PersistentMenuFivePayload):
                    return ReceivingType.PERSISTENT_MENU_FIVE
                elif isinstance(subpayload, GetStartedPayload):
                    return ReceivingType.GET_STARTED_BUTTON
                else:
                    return ReceivingType.POSTBACK
            except Exception as e:
                return ReceivingType.POSTBACK


class AuthenticationChecker(MessagingChecker):
    def check(self):
        if not hasattr(self.messaging, 'optin'):
            return None
        if self.messaging.optin is not None and self.messaging.optin.ref:
            return ReceivingType.AUTHENTICATION


class AccountLinkingChecker(MessagingChecker):
    def check(self):
        if not hasattr(self.messaging, 'account_linking'):
            return None
        if self.messaging.account_linking is not None \
                and self.messaging.account_linking.status:
            return ReceivingType.ACCOUNT_LINKING


class MessageDeliveredChecker(MessagingChecker):
    def check(self):
        if not hasattr(self.messaging, 'delivery'):
            return None
        if self.messaging.delivery is not None and self.messaging.delivery.watermark:
            return ReceivingType.DELIVERED


class MessageReadChecker(MessagingChecker):
    def check(self):
        if not hasattr(self.messaging, 'read'):
            return None
        if self.messaging.read is not None and self.messaging.read.watermark:
            return ReceivingType.READ


class Sender(Serializable):
    """
    In Receiving messages, Sender represents the Facebook user who sent this
    message to your page
    """
    def __init__(self, iid):
        """
        When representing a user, these IDs are page-scoped IDs (PSID).
        This means that the IDs of users are unique for a given page.

        @param iid: user's PSID
        @type iid: str
        """
        self.id = iid


class Recipient(Serializable):
    """
    Your page
    """
    def __init__(self, iid):
        """
        @param iid: Facebook page id
        @type iid: str
        """
        self.id = iid


class Read(Serializable):
    def __init__(self, watermark, seq):
        """

        @param watermark:  All messages that were sent before this timestamp
                           were read
        @type watermark: int

        @param seq: Sequence number
        @type seq: int
        """
        self.watermark = watermark
        self.seq = seq


class Delivery(Serializable):
    def __init__(self, watermark, seq, mids=[]):
        """

        @param watermark: All messages that were sent before this timestamp
                          were delivered
        @type watermark: int

        @param seq: Sequence number
        @type seq: int

        @param mids: Array containing message IDs of messages that were
                     delivered. Field may not be present.
        @type mids: list of str

        """
        self.watermark = watermark
        self.seq = seq
        self.mids = mids


class AccountLinking(Serializable):

    def __init__(self, status, authorization_code=""):
        """

        @param status: linked or unlinked
        @type status: str

        @param authorization_code: Value of pass-through authorization_code
                                   provided in the Linking Account flow
        @type status: str
        """
        self.status = status
        self.authorization_code = authorization_code


class Optin(Serializable):
    def __init__(self, ref):
        """

        @param ref: data-ref parameter that was defined with the entry point
        @type ref: str
        """
        self.ref = ref


class Postback(Serializable):
    def __init__(self, payload):
        """

        @param payload:  payload parameter that was defined with the button
        @type payload: str
        """
        self.payload = payload


class QuickReply(Serializable):
    def __init__(self, payload):
        """

        @param payload: Custom data provided by the app
        @type payload: str
        """
        self.payload = payload


class MultimediaAttachmentPayload(Serializable):
    def __init__(self, url):
        """

        @param url: URL of the file
        @type url: str
        """
        self.url = url


class Coordinates(Serializable):
    def __init__(self, lat, llong):
        """
        lat and llong must be convertible to Decimal

        @param lat: latitude
        @type lat: number

        @param llong: longitude
        @type llong: number
        """
        self.lat = Decimal(lat)
        self.long = Decimal(llong)


class LocationAttachmentPayload(Serializable):
    custom_obj_map = {
        'coordinates': [Coordinates, object]
    }

    def __init__(self, coordinates):
        """

        @param coordinates:
        @type coordinates: Coordinates
        """
        self.coordinates = coordinates


class AttachmentType(SerializableEnum):
    pass


AttachmentType.IMAGE = AttachmentType('image')
AttachmentType.AUDIO = AttachmentType('audio')
AttachmentType.VIDEO = AttachmentType('video')
AttachmentType.FILE = AttachmentType('file')
AttachmentType.LOCATION = AttachmentType('location')


class Attachment(Serializable):
    """
    There are 2 types of Attachment payloads, LocationAttachmentPayload and
    MultimediaAttachmentPayload, so we are not providing _obj_map but need to
    override deserialize to look at self.type and determine which type of
    payload to deserialize to
    """
    def __init__(self, ttype, payload, url):
        """

        @param ttype:
        @type ttype: str
        @param payload: either MultimediaAttachmentPayload or LocationAttachmentPayload
        @type payload: MultimediaAttachmentPayload
        @type payload: LocationAttachmentPayload
        """
        self.type = ttype
        self.payload = payload
        self.url = url

    @classmethod
    def deserialize(cls, json_data):
        # FYI: 2 attachment types, so need to check on self.type to see which
        # one to use
        if 'type' not in json_data or 'payload' not in json_data:
            raise ValueError('Deserializing Attachment must have type and '
                             'payload in the json data but was missing')
        try:
            ttype = AttachmentType.value_of(json_data['type'])
        except:
            raise ValueError('Given type is not defined: %s' % json_data['type'])
        if ttype == AttachmentType.LOCATION:
            payload_clazz = LocationAttachmentPayload
        else:
            payload_clazz = MultimediaAttachmentPayload
        url = getattr(json_data, 'url', None)
        return cls(json_data['type'],
                   payload_clazz.deserialize(json_data['payload']),
                   url
                   )


class Message(Serializable):
    """
    Each Messaging could have 1 Message object
    """
    custom_obj_map = {
        'quick_reply': [QuickReply, object],
        'attachments': [Attachment, list]
    }

    def __init__(self, text, mid, seq, sticker_id=None, quick_reply=None, attachments=[],
                 is_echo=False, app_id="", metadata=""):
        """

        @param text: text of message
        @type text: str

        @param mid: message ID
        @type mid: str

        @param seq: message sequence number
        @type seq: int

        @param quick_reply: optional custom data provided by the sending app
        @type quick_reply: QuickReply

        @param attachments: Array containing attachment data
        @type attachments: list(Attachment)

        @param is_echo: Indicates the message sent from the page itself
        @type is_echo: bool

        @param app_id: ID of the app from which the message was sent
        @type app_id: str

        @param metadata: Custom string passed to the Send API as the metadata field
        @type metadata: str

        @param sticker_id: if user sends a sticker like thumb up
        @type sticker_id: int
        """
        self.text = text
        self.mid = mid
        self.seq = seq
        self.quick_reply = quick_reply
        self.attachments = attachments
        self.is_echo = is_echo
        self.app_id = app_id
        self.metadata = metadata
        self.sticker_id = sticker_id


class Messaging(Serializable):
    """
    Each Entry contains multiple Messaging object, Messaging has different
    CallbackType, we are going to define this class so that it includes all
    possible objects and then when we got a json data, we just load them
    into the objects. After we deserialized json data into Messaging, we then
    determine what type of a Callback this is by looking at some fields and
    values using CallbackTypeChecker
    """
    custom_obj_map = {
        'sender': [Sender, object],
        'recipient': [Recipient, object],
        'message': [Message, object],
        'postback': [Postback, object],
        'optin': [Optin, object],
        'account_linking': [AccountLinking, object],
        'delivery': [Delivery, object],
        'read': [Read, object]
    }

    def __init__(self, sender, recipient, timestamp=int(time.time()),
                 message=None, postback=None, optin=None, account_linking=None,
                 delivery=None, read=None):
        """

        @param sender: Sender object
        @type sender: Sender

        @param recipient: Recipient object
        @type recipient: Recipient

        @param timestamp: epoch
        @type timestamp: int

        @param message: message block
        @type message: Messaging

        @param postback: postback block
        @type postback: Postback

        @param optin: optin block
        @type optin: Optin

        @param account_linking: account_linking block
        @type account_linking: AccountLinking

        @param delivery: delivery block
        @type delivery: Delivery

        @param read: read block
        @type read: Read
        """
        self.sender = sender
        self.recipient = recipient
        self.timestamp = timestamp
        self.message = message
        self.postback = postback
        self.optin = optin
        self.account_linking = account_linking
        self.delivery = delivery
        self.read = read
        self._receiving_type = None

    def __repr__(self):
        return 'Messaging sent from %s ReceivingType = %s' \
               % (self.get_psid(), self.get_receiving_type())

    def __str__(self):
        return repr(self)

    def get_receiving_type(self):
        """
        After loading the json data into this object, try to determine what
        type of callback this messaging is

        @return:
        @rtype: ReceivingType
        """
        if hasattr(self, '_callback_type') and self._receiving_type:
            return self._receiving_type
        # we want to first see if messaging is a MessageReceived
        res = MessageReceivedChecker(self).check()
        # loop through all subclasses of MessagingCheck and if any of it passed
        # return it
        if res:
            self._receiving_type = res
        else:
            for clazz in MessagingChecker.__subclasses__():
                res = clazz(self).check()
                if res:
                    self._receiving_type = res
        return self._receiving_type

    def get_text(self):
        """

        @return:
        @rtype: str
        """
        if hasattr(self, 'message') and hasattr(self.message, 'text') \
                and self.message.text:
            return self.message.text

    def get_url(self):
        """

        @return:
        @rtype: str
        """
        if hasattr(self, 'message') \
                and hasattr(self.message, 'attachments') \
                and len(self.message.attachments) > 0 \
                and hasattr(self.message.attachments[0], 'payload') \
                and hasattr(self.message.attachments[0].payload, 'url'):
            return self.message.attachments[0].payload.url

    def get_coordinates(self):
        """

        @return:
        @rtype: Coordinates
        """
        if hasattr(self, 'message') \
                and hasattr(self.message, 'attachments') \
                and len(self.message.attachments) > 0 \
                and hasattr(self.message.attachments[0], 'payload') \
                and hasattr(self.message.attachments[0].payload, 'coordinates'):
            return self.message.attachments[0].payload.coordinates

    def get_postback_payload(self):
        """

        @return:
        @rtype: str
        """
        if hasattr(self, 'postback') \
                and hasattr(self.postback, 'payload'):
            return self.postback.payload

    def get_quick_reply_payload(self):
        if hasattr(self, 'message') \
                and hasattr(self.message, 'quick_reply') \
                and hasattr(self.message.quick_reply, 'payload'):
            return self.message.quick_reply.payload

    def get_watermark(self):
        pass

    def get_psid(self):
        if hasattr(self, 'sender') \
                and hasattr(self.sender, 'id') \
                and self.sender.id:
            return self.sender.id

    def get_sticker_id(self):
        if hasattr(self, 'message') \
                and hasattr(self.message, 'sticker_id') \
                and self.message.sticker_id:
            return self.message.sticker_id


class Entry(Serializable):
    """
    https://developers.facebook.com/docs/messenger-platform/webhook-reference

    """
    custom_obj_map = {
        'messaging': [Messaging, list]
    }

    def __init__(self, iid, time, messaging=[]):
        """

        @param iid: page id
        @type iid: str

        @param time: time of update (epoch time in milliseconds)
        @type time: int

        @param messaging: Array of Messaging
        @type messaging: list(Messaging)
        """
        self.id = iid
        self.time = time
        # messaging is a list of Messaging object
        self.messaging = messaging


class Callback(Serializable):
    """
    https://developers.facebook.com/docs/messenger-platform/webhook-reference
    """
    custom_obj_map = {
        'entry': [Entry, list]
    }

    def __init__(self, object='page', entry=[]):
        """

        @param object: must be "page"
        @type object: str

        @param entry: list of Entry
        @type entry: list(Entry)
        """
        self.object = object
        # entry is a list of Entry objects
        self.entry = entry

    def get_sender_id(self):
        """
        Each callback must have some Messaging, we should expect that we can
        always get a sender id

        @return:
        """
        ret = None
        for entry in self.entry:
            for messaging in entry.messaging:
                if hasattr(messaging, 'sender') \
                        and hasattr(messaging.sender, 'id') \
                        and messaging.sender.id:
                    return messaging.sender.id



# -*- coding: utf-8 -*-
import importlib
import logging


logger = logging.getLogger(__name__)


def load_class(fullyname):
    """Load class by its fully qualified name

    @param fullyname: if given fully qualified name is a.b.c, try to load the
                      module from a.b and load the class c in a.b module
    @return: A class
    @raise ImportError: not able to load module
    @raise AttributeError: not able to load class
    """
    tokens = fullyname.split('.')
    # if len(tokens) == 1, then it is a module, raise exception
    if len(tokens) == 1:
        raise ValueError('The fully qualified name you gave was %s which can '
                         'not represent a class but a module' % fullyname)
    else:
        class_name = tokens.pop()
        module_name = '.'.join(tokens)

    try:
        module = importlib.import_module(module_name)
    except ImportError as e:
        logger.error('Not able to import module named %s' % module_name)
        raise e
    try:
        ret = getattr(module, class_name)
    except AttributeError as e:
        logger.error('Not able to get class %s from module %s' %
                     (class_name, module_name))
        raise e
    return ret


def get_class_name(obj):
    """
    Given a type of an instance, return its fully qualified class name

    @param obj:
    @type obj: type or object
    @return: fully qualified class name
    @rtype: str
    """
    if not isinstance(obj, type):
        raise ValueError('Given class must be a type')
        return '%s.%s' % (obj.__module__, obj.__name__)
    else:
        return '%s.%s' % (obj.__module__, obj.__class__.__name__)



# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2016-08-24 06:49
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djmessenger', '0009_auto_20160824_0625'),
    ]

    operations = [
        migrations.AlterModelTable(
            name='getstartedbutton',
            table='djm_getstarted',
        ),
    ]



# -*- coding: utf-8 -*-
import requests
import logging


from django.core.management.base import BaseCommand

from djmessenger.settings import DJM_THREAD_SETTINGS_URL, DJM_THREAD_GREETINGS_TEXT


logger = logging.getLogger(__name__)


class Command(BaseCommand):
    def handle(self, *args, **options):
        logger.debug('Ready to configure Greetings')
        data = {
            "setting_type": "greeting",
            "greeting": {
                "text": str(DJM_THREAD_GREETINGS_TEXT)
            }
        }
        logger.debug('Prepared payload for configuring Greetings as %s'
                     % data)
        status = requests.post(DJM_THREAD_SETTINGS_URL,
                               headers={"Content-Type": "application/json"},
                               json=data)
        print(status.content)
        logger.debug('Configuring greetings got response code %s, content %s'
                     % (status.status_code, status.content))




# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2016-08-23 22:43
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('djmessenger', '0002_callback_fbuserprofile_getstartedbutton_persistentmenu'),
    ]

    operations = [
        migrations.CreateModel(
            name='UserLocation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('latitude', models.DecimalField(blank=True, decimal_places=8, max_digits=10, null=True)),
                ('longitude', models.DecimalField(blank=True, decimal_places=8, max_digits=11, null=True)),
                ('timestamp', models.BigIntegerField(help_text='Facebook returns timestamp as EPOCH')),
                ('url', models.URLField(blank=True, max_length=1000, null=True)),
                ('mid', models.CharField(blank=True, max_length=512, null=True)),
                ('seq', models.IntegerField(blank=True, null=True)),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='djmessenger.FBUserProfile')),
            ],
            options={
                'get_latest_by': 'date_created',
            },
        ),
    ]
